# Copyright (C) 2025 Cardiff University
# SPDX-License-Identifier: MIT

"""Core utilities for IGWN Robot Auth tools."""

from __future__ import annotations

import argparse
import logging

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# -- log formatting

DEFAULT_FORMAT = "%(asctime)s:%(name)s[%(process)d]:%(levelname)s:%(message)s"
DEFAULT_DATEFMT = "%Y-%m-%dT%H:%M:%S%z"


def configure_logging(
    verbose: int = 0,
    fmt: str = DEFAULT_FORMAT,
    datefmt: str = DEFAULT_DATEFMT,
    *,
    force: bool = False,
    **kwargs,
) -> None:
    """Configure a `logging.Logger` for a command-line application.

    Parameters
    ----------
    verbose : `int`
        Verbosity level (0-3).

    fmt : `str`
        Log message format string.

    datefmt : `str`
        Date format string.

    force : `bool`, optional
        If `True`, force reconfiguration of logging even if the root logger is
        already configured.

    kwargs
        Additional keyword arguments passed to the logging configuration.

    Notes
    -----
    If the Python root logger is already configured, this function will do
    nothing unless `force` is set to `True`.
    """
    if logging.getLogger().hasHandlers() and not force:
        return
    level = max(3 - verbose, 0) * 10
    try:
        import coloredlogs
    except ImportError:
        logging.basicConfig(
            format=fmt,
            level=level,
            datefmt=datefmt,
            **kwargs,
        )
    else:
        coloredlogs.install(
            fmt=fmt,
            level=level,
            datefmt=datefmt,
            **kwargs,
        )


# -- argument parsing

class ArgumentFormatter(
    argparse.RawDescriptionHelpFormatter,
    argparse.ArgumentDefaultsHelpFormatter,
):
    """`argparse.ArgumentFormatter` for `igwn-robot-auth` tools."""


class ArgumentParser(argparse.ArgumentParser):
    """`argparse.ArgumentParser` for `igwn-robot-auth` tools."""

    def __init__(
        self,
        version: str | None = None,
        man_short_description: str | None = None,
        manpage: list[dict[str, str]] | None = None,
        **kwargs,
    ) -> None:
        """Initialise the `ArgumentParser`."""
        # initialise ArgumentParser
        kwargs.setdefault("formatter_class", ArgumentFormatter)
        super().__init__(**kwargs)

        # update argument group titles
        self._positionals.title = "Required (positional) arguments"
        self._optionals.title = "Optional arguments"

        # handle formatting for argparse-manpage
        self._manpage = manpage
        if not man_short_description and self.description:
            # if man_short_description was not given, copy the first line of
            # the parser description, with some pedantic reformatting
            man_short_description = (
                self.description.split("\n")[0].lower().rstrip(".")
            )
        self.man_short_description = man_short_description

        # add default -V/--version
        if version is not False:
            from igwn_robot_auth import __version__
            self.add_argument(
                "-V",
                "--version",
                action="version",
                version=version or __version__,
            )



# Copyright (c) 2024-2025 Cardiff University

import os
from datetime import (
    datetime,
    timezone,
)

import sphinx_github_style

from igwn_robot_auth import __version__ as igwn_robot_auth_version

# -- Project information -------------

project = "igwn-robot-auth"
copyright = f"{datetime.now(tz=timezone.utc).date().year}, Cardiff University"
author = "Duncan Macleod"

# The full version, including alpha/beta/rc tags
release = igwn_robot_auth_version
version = release.split("+", 1)[0]

# The version used in the image tag.
# This is used to provide an accurate tag for the Docker image docs
if os.getenv("READTHEDOCS_VERSION") == "latest":
    image_tag = os.getenv("READTHEDOCS_GIT_COMMIT_HASH", version)
elif os.getenv("GITLAB_CI"):
    image_tag = os.getenv("CI_COMMIT_TAG", os.getenv("CI_COMMIT_SHA", version))
else:
    image_tag = version

# -- General configuration -----------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Content to prepend to all reStructuredText files.
# This is used to support substitutions like |release| and |version|.
rst_prolog = f"""
.. |image_name| replace:: containers.ligo.org/computing/software/igwn-robot-auth
.. |image_tag| replace:: {image_tag}
"""

# -- Options for HTML output ---------

html_theme = "furo"
html_title = f"{project} {version}"

html_static_path = ["_static"]
html_css_files = ["igwn-furo.css"]

pygments_dark_style = "monokai"

default_role = "obj"

maximum_signature_line_length = 80

# -- Extensions ----------------------

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_automodapi.automodapi",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_substitution_extensions",
    "sphinxarg.ext",
]

# -- automodapi

automodapi_inherited_members = False

# -- copybutton

copybutton_prompt_text = " |".join((  # noqa: FLY002
    ">>>",
    r"\.\.\.",
    r"\$"
    r"In \[\d*\]:",
    r" {2,5}\.\.\.:",
    " {5,8}: ",
))
copybutton_prompt_is_regexp = True
copybutton_line_continuation_character = "\\"

# -- intersphinx

intersphinx_mapping = {name: (url, None) for name, url in {
    "htcondor": "https://htcondor.readthedocs.io/en/latest/",
    "igwn-auth-utils": "https://igwn-auth-utils.readthedocs.io/en/stable/",
    "python": "https://docs.python.org/3/",
    "requests": "https://requests.readthedocs.io/en/stable/",
    "scitokens": "https://scitokens.readthedocs.io/en/latest/",
}.items()}

# -- linkcode

linkcode_url = sphinx_github_style.get_linkcode_url(
    blob=sphinx_github_style.get_linkcode_revision("head"),
    url=f"https://git.ligo.org/computing/software/{project}",
)
linkcode_resolve = sphinx_github_style.get_linkcode_resolve(linkcode_url)

# -- myst_parser

myst_enable_extensions = [
    "attrs_block",
]

# -- napoleon

napoleon_use_rtype = False



# Copyright 2025 Cardiff University
# SPDX-License-Identifier: MIT

"""Get credentials for an IGWN robot."""

from __future__ import annotations

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import logging
import os
import typing
from datetime import (
    datetime,
    timezone,
)
from pathlib import Path
from unittest import mock

from igwn_auth_utils import (
    get_scitoken,
    kinit,
)
from igwn_auth_utils.scitokens import (
    default_bearer_token_file,
    load_token_file,
)

from igwn_robot_auth.htcondor import condor_vault_storer
from igwn_robot_auth.tools.core import (
    ArgumentParser,
    configure_logging,
)

if typing.TYPE_CHECKING:
    from scitokens import SciToken

log = logging.getLogger(__name__)

DEFAULT_BEARER_TOKEN_FILE = default_bearer_token_file()
DEFAULT_MINSECS = 3600
DEFAULT_REALM = "LIGO.ORG"
DEFAULT_VAULT_SERVER = "vault.ligo.org"
DEFAULT_SCITOKEN_ISSUER = "igwn"
KEYTAB_ENVIRONMENT_VARIABLE = "KRB5_KTNAME"

MANPAGE = [
    {
        "heading": "environment",
        "content": f"""
.TP
.B "BEARER_TOKEN_FILE"
The default path for the SciToken
.TP
.B "{KEYTAB_ENVIRONMENT_VARIABLE}"
The default path for the Kerberos keytab
""",
    },
]


def format_principal(
    principal: str,
    default_realm: str = DEFAULT_REALM,
) -> str:
    """Format a Kerberos principal with a default realm."""
    if "@" in principal:
        return principal
    return f"{principal}@{default_realm}"


def default_credkey(
    principal: str,
) -> str:
    """Return the default `--credkey` argument based on the principal."""
    return principal.split("@", 1)[0]


def default_role(
    principal: str,
) -> str:
    """Return the default `--role` argument based on the principal."""
    return principal.split("/", 1)[0]


def debug_token(token: SciToken) -> None:
    """Display this tokens claims as ``DEBUG``-level log messages."""
    log.debug("Token claims:")
    for claim, value in sorted(token.claims()):
        if isinstance(value, int) and claim in {"exp", "iat", "nbf"}:
            value = datetime.fromtimestamp(value, timezone.utc)  # noqa: PLW2901
        log.debug("%8s: %s", claim, value)


# -- Python entry point

def get(
    principal: str | None = None,
    keytab: str | None = None,
    outfile: str = DEFAULT_BEARER_TOKEN_FILE,
    issuer: str = DEFAULT_SCITOKEN_ISSUER,
    ccache: str | None = None,
    credkey: str | None = None,
    role: str | None = None,
    minsecs: int = DEFAULT_MINSECS,
    vaultserver: str = DEFAULT_VAULT_SERVER,
    vaulttokenfile: str | None = None,
    vaulttokenminttl: str | int = "24h",
    *,
    condor: bool = False,
) -> None:
    """Get a token for a robot Kerberos principal.

    Parameters
    ----------
    principal : `str`
        Principal name for Kerberos credential.
        If not given it will be taken from the ``keytab``.
        If ``principal`` is not specified in the form ``name@REALM``
        the default realm REALM will be applied, see ``man krb5.conf``.

    keytab : `str`, optional
        Path to keytab file.
        Default taken from ``KRB5_KTNAME`` environment variable.
        If the environment variable is not set, or is empty, this keyword is
        required.

    outfile : `str`, optional
        Path in which to write the serialised `~scitokens.SciToken`.

    issuer : `str`, optional
        Name of vault token issuer.
        Default is ``"igwn"``.

    ccache : `str`, optional
        Path to Kerberos credentials cache.
        Default is the default credential cache, see ``man krb5.conf``.

    credkey : `str`, optional
        Vault credential key for this identity.
        Default is derived from Kerberos principal.

    role : `str`, optional
        Vault name of role for this identity.
        Default is derived from Kerberos principal.

    minsecs : `int`, optional
        Minimum number of seconds left in bearer token before expiry.
        If an existing token is found with a remaining lifetime greater
        than this number, `htgettoken` will not renew it.

    vaultserver : `str`, optional
        Name or IP of vault server to use

    vaulttokenfile : `str`, optional
        Path in which to store/use vault token.

    vaulttokenminttl : `str`, optional
        Minimum remaining lifetime of vault token before attempting renewal.

    condor : `bool`, optional
        Use `condor_vault_storer` to initialise a token for HTCondor.
        Default is `False`.

    Examples
    --------
    Get a SciToken for a specific robot keytab, dynamically grabbing the
    principal name from the keytab:

    >>> get(keytab="/home/user/.secure/robot.keytab")

    See Also
    --------
    igwn_auth_utils.kinit
        For details of how Kerberos credentials are initialised.

    igwn_auth_utils.get_scitoken
        For details of how a `~scitoken.SciToken` is acquired.

    igwn_robot_auth.htcondor.condor_vault_storer
        For details of how ``condor_vault_storer`` is invoked when
        ``condor=True`` is given.
    """
    # -- step 1: get a Kerberos ticket-granting-ticket

    log.info("Getting Kerberos credential")
    creds = kinit(
        principal,
        keytab=keytab,
        ccache=ccache,
    )
    principal = str(creds.name)
    log.debug("Acquired Kerberos credential for %s", principal)

    # -- step 2: get a SciToken using Kerberos

    if credkey is None:
        credkey = default_credkey(principal)
        log.debug("Set default --credkey='%s'", credkey)
    if role is None:
        role = default_role(principal)
        log.debug("Set default --role='%s'", role)

    ccacheenv = {}
    if ccache:
        ccacheenv["KRB5CCNAME"] = ccache

    with mock.patch.dict("os.environ", ccacheenv):
        # get a token
        if condor:
            condor_vault_storer(
                issuer,
                credkey=credkey,
                role=role,
                vaultserver=vaultserver,
            )
        else:
            log.info("Getting SciToken")
            tokenfile = get_scitoken(
                credkey=credkey,
                issuer=issuer,
                outfile=outfile,
                role=role,
                minsecs=minsecs,
                vaultserver=vaultserver,
                vaulttokenfile=vaulttokenfile or False,
                vaulttokenminttl=vaulttokenminttl,
                verbose=log.isEnabledFor(logging.INFO),
                debug=log.isEnabledFor(logging.DEBUG),
            )

            # debug the token
            if log.isEnabledFor(logging.DEBUG):
                debug_token(load_token_file(tokenfile))

    return outfile


# -- command-line options

def create_parser() -> ArgumentParser:
    """Create an `argparse.ArgumentParser` for this tool."""
    parser = ArgumentParser(
        description=__doc__,
        manpage=MANPAGE,
    )

    # -- positionals

    parser.add_argument(
        "principal",
        action="store",
        type=format_principal,
        nargs="?",
        help="Kerberos principal name (required if -k/--keytab not given)",
    )

    # -- options

    # general
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increate verbosity (can be given multiple times).",
    )

    # kerberos
    kerberosargs = parser.add_argument_group(
        "Kerberos options",
    )
    kerberosargs.add_argument(
        "-k",
        "--keytab",
        "--kerberos-keytab",
        action="store",
        default=os.getenv(KEYTAB_ENVIRONMENT_VARIABLE),
        required=not os.getenv(KEYTAB_ENVIRONMENT_VARIABLE),
        type=Path,
        help="Path to Kerberos keytab file",
    )
    kerberosargs.add_argument(
        "-c",
        "--ccache",
        "--kerberos-ccache",
        help="Path to Kerberos ccache (default is default Kerberos ccache)",
    )

    # scitokens
    tokenargs = parser.add_argument_group(
        "SciToken options",
    )
    tokenargs.add_argument(
        "-a",
        "--vaultserver",
        default=DEFAULT_VAULT_SERVER,
        help="Name or IP of vault server to use",
    )
    tokenargs.add_argument(
        "-i",
        "--issuer",
        default=DEFAULT_SCITOKEN_ISSUER,
        help="Name of SciToken issuer",
    )
    tokenargs.add_argument(
        "--credkey",
        help=(
            "Vault credential key for this identity; "
            "default is derived from Kerberos principal"
        ),
    )
    tokenargs.add_argument(
        "-r",
        "--role",
        help=(
            "Vault name of role for this identity; "
            "default is derived from Kerberos principal"
        ),
    )
    tokenargs.add_argument(
        "-m",
        "--minsecs",
        type=int,
        default=DEFAULT_MINSECS,
        help=(
            "Minimum number of seconds left in bearer token before expiry; "
            "if an existing token is found with a remaining lifetime greater "
            "than this number, htgettoken will not renew it"
        ),
    )
    tokenargs.add_argument(
        "--vaulttokenfile",
        default=None,
        help="Path in which to store/use vault token",
    )
    tokenargs.add_argument(
        "-o",
        "--outfile",
        "--bearertokenfile",
        default=DEFAULT_BEARER_TOKEN_FILE,
        help="Path in which to store bearer token",
    )
    tokenargs.add_argument(
        "--vaulttokenminttl",
        default="24h",
        help="Minimum remaining lifetime of vault token before attempting renewal",
    )

    condorargs = parser.add_argument_group(
        "Condor options",
    )
    condorargs.add_argument(
        "--condor",
        action="store_true",
        default=False,
        help="Use `condor_vault_storer` to initialise a token for HTCondor",
    )
    return parser


# -- main entry point

def main(args: list[str] | None = None) -> None:
    """Run this tool.

    Parameters
    ----------
    args : `list` of `str`
        The list of arguments to parse.
        Passed directly to `argparse.ArgumentParser.parse_args`.

    configure_logger : `bool`, optional
        Whether to configure the root logger.
        Default is `True`.
        If calling this function from another application or library, you
        probably want to pass ``configure_logger=False`` to allow the end
        application/user to configure logging themselves.
    """
    parser = create_parser()
    opts = parser.parse_args(args=args)

    # set verbose logging
    configure_logging(verbose=opts.verbose)

    # get the creds
    get(
        principal=opts.principal,
        keytab=opts.keytab,
        outfile=opts.outfile,
        issuer=opts.issuer,
        ccache=opts.ccache,
        credkey=opts.credkey,
        role=opts.role,
        minsecs=opts.minsecs,
        vaultserver=opts.vaultserver,
        vaulttokenfile=opts.vaulttokenfile,
        vaulttokenminttl=opts.vaulttokenminttl,
        condor=opts.condor,
    )


if __name__ == "__main__":
    main()



# Copyright (c) 2025 Cardiff University
# SPDX-License-Identifier: MIT

"""Tests for `igwn_robot_auth.tools.get`."""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import pytest

from igwn_robot_auth.tools import get as tools_get


def test_help():
    """Test the ``--help`` option."""
    with pytest.raises(SystemExit) as excinfo:
        tools_get.main(["--help"])
    assert excinfo.value.code == 0



# Copyright (c) 2025 Cardiff University
# SPDX-License-Identifier: MIT

"""Auth credential management for IGWN robots.

IGWN Robot Auth provides utilities to help manage authorisation credentials
for automated processes run by the `LIGO <https://www.ligo.org>`__
- `Virgo <https://www.virgo-gw.eu>`__
- `KAGRA <https://gwcenter.icrr.u-tokyo.ac.jp>`__ collaborations.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

from .tools.get import get

try:
    from ._version import version as __version__
except ModuleNotFoundError:  # development mode
    __version__ = "dev"



# Copyright (c) 2025 Cardiff University
# SPDX-License-Identifier: MIT

"""Manage authorisation credentials for IGWN Robot identities."""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

from .tools.get import main as get_main

if __name__ == "__main__":
    get_main()



# Copyright (c) 2025 Cardiff University
# SPDX-License-Identifier: MIT

"""Tests for `igwn_robot_auth`."""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

from pathlib import Path
from unittest import mock

import pytest

from igwn_robot_auth import htcondor as ira_condor


@mock.patch("igwn_robot_auth.htcondor.Path.home", return_value=Path("/home/user"))
def test_condor_config_path(mock_home):
    """Test `condor_config_path()`."""
    assert ira_condor.condor_config_path() == Path(
        "/home/user/.condor/user_config",
    )


@mock.patch(
    "igwn_robot_auth.htcondor._get_config",
    return_value="--one=1 --two=2 --three=3",
)
def test_check_condor_config(mock_condor_config_path):
    """Test `check_config_config()`."""
    ira_condor.check_condor_config(one=1, two=2, three=3)


@mock.patch(
    "igwn_robot_auth.htcondor._get_config",
    return_value="--one=1 --two=2 --three=3",
)
@mock.patch("igwn_robot_auth.htcondor.condor_config_path")
@pytest.mark.parametrize(("exists", "match"), [
    (False, "No HTCondor user_config file found"),
    (True, "HTCondor user_config value for SEC_CREDENTIAL_GETTOKEN_OPTS"),
])
def test_check_condor_config_error(
    mock_condor_config_path,
    mock_get_config,
    exists,
    match,
):
    """Test `check_config_config()` error handling."""
    mock_condor_config_path.return_value.exists.return_value = exists
    with pytest.raises(ValueError, match=match):
        ira_condor.check_condor_config(one=2)



# Copyright (c) 2025 Cardiff University
# SPDX-License-Identifier: MIT

"""HTCondor utilities for IGWN Robot Auth."""

from __future__ import annotations

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import logging
import shlex
import shutil
import subprocess
from pathlib import Path

log = logging.getLogger(__name__)

CONDOR_GETTOKEN_OPTS_PARAM = "SEC_CREDENTIAL_GETTOKEN_OPTS"
CONDOR_USER_CONFIG_FILE_PARAM = "USER_CONFIG_FILE"
CONDOR_VAULT_STORER = shutil.which("condor_vault_storer") or "condor_vault_storer"


def _get_config(param: str) -> str:
    """Get a parameter from the HTCondor config."""
    import htcondor
    value = htcondor.param[param]
    log.debug("htcondor.param: %s = %s", param, value)
    return value


def condor_config_path() -> Path:
    """Return the path of the HTCondor user config."""
    user_config = _get_config(CONDOR_USER_CONFIG_FILE_PARAM)
    return Path.home() / ".condor" / user_config


def _match_config(args: list[str]) -> None:
    """Match the current HTCondor configuration against the given ``args``."""
    found = _get_config(CONDOR_GETTOKEN_OPTS_PARAM)
    foundargs = set(shlex.split(found))
    if foundargs.symmetric_difference(args):
        argstr = shlex.join(args)
        msg = (
           f"HTCondor user_config value for {CONDOR_GETTOKEN_OPTS_PARAM} "
           f"in {condor_config_path()} doesn't match current options: "
           f"{found!r} vs {argstr!r}"
        )
        raise ValueError(msg)


def check_condor_config(
    **options,
) -> None:
    """Check that the HTCondor user config matches the specified options."""
    args = [f"--{key}={value}" for key, value in options.items()]
    try:
        _match_config(args)
    except ValueError as exc:
        config = condor_config_path()
        if not config.exists():
            msg = f"""No HTCondor user_config file found at {config}.
Please consider creating one with the following contents:

{CONDOR_GETTOKEN_OPTS_PARAM} = {' '.join(args)}
"""
            raise ValueError(msg) from exc
        raise


def condor_vault_storer(
    oauth_service: str = "igwn",
    *,
    verbose: bool = True,
    **htgettoken_opts,
) -> None:
    """Call out to ``condor_vault_storer`` with the given options.

    This function just runs `subprocess.check_call`.

    Parameters
    ----------
    oauth_service : `str`, optional
        The ``oauth_service`` argument to use.
        This is typically just the name of the token issuer.
        Default is ``"igwn"``.

    verbose : `bool`, optional
        Show verbose output.

    htgettoken_opts
        Other ``key=value`` arguments are passed verbatim to the
        ``condor_vault_storer`` call.
    """
    check_condor_config(**htgettoken_opts)
    log.debug("Condor user_config matches current options")
    verbosearg = ("-v",) if verbose else ()
    log.debug("Executing condor_vault_storer")
    cmd = [
        CONDOR_VAULT_STORER,
        *verbosearg,
        oauth_service,
    ]
    cmdstr = shlex.join(cmd)
    log.debug("$ %s", cmdstr)
    try:
        subprocess.check_call(cmd, shell=False)
    except subprocess.CalledProcessError as exc:
        exc.cmd = cmdstr
        raise
